import logging
from functools import cached_property

from .grocy_api_client import DEFAULT_PORT_NUMBER, GrocyApiClient
from .managers.batteries import BatteryManager
from .managers.calendar import CalendarManager
from .managers.chores import ChoreManager
from .managers.equipment import EquipmentManager
from .managers.files import FileManager
from .managers.generic import GenericEntityManager
from .managers.meal_plan import MealPlanManager
from .managers.recipes import RecipeManager
from .managers.shopping_list import ShoppingListManager
from .managers.stock import StockManager
from .managers.system import SystemManager
from .managers.tasks import TaskManager
from .managers.users import UserManager

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class Grocy:
    def __init__(
        self,
        base_url,
        api_key,
        port: int = DEFAULT_PORT_NUMBER,
        path: str | None = None,
        verify_ssl=True,
        debug=False,
    ):
        self._api_client = GrocyApiClient(
            base_url, api_key, port, path, verify_ssl, debug
        )

        if debug:
            _LOGGER.setLevel(logging.DEBUG)

    @cached_property
    def stock(self) -> StockManager:
        return StockManager(self._api_client)

    @cached_property
    def shopping_list(self) -> ShoppingListManager:
        return ShoppingListManager(self._api_client)

    @cached_property
    def recipes(self) -> RecipeManager:
        return RecipeManager(self._api_client)

    @cached_property
    def chores(self) -> ChoreManager:
        return ChoreManager(self._api_client)

    @cached_property
    def tasks(self) -> TaskManager:
        return TaskManager(self._api_client)

    @cached_property
    def batteries(self) -> BatteryManager:
        return BatteryManager(self._api_client)

    @cached_property
    def equipment(self) -> EquipmentManager:
        return EquipmentManager(self._api_client)

    @cached_property
    def meal_plan(self) -> MealPlanManager:
        return MealPlanManager(self._api_client)

    @cached_property
    def users(self) -> UserManager:
        return UserManager(self._api_client)

    @cached_property
    def system(self) -> SystemManager:
        return SystemManager(self._api_client)

    @cached_property
    def generic(self) -> GenericEntityManager:
        return GenericEntityManager(self._api_client)

    @cached_property
    def calendar(self) -> CalendarManager:
        return CalendarManager(self._api_client)

    @cached_property
    def files(self) -> FileManager:
        return FileManager(self._api_client)



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel


class Battery(BaseModel):
    id: int
    name: str | None = None
    description: str | None = None
    used_in: str | None = None
    charge_interval_days: int | None = None
    created_timestamp: datetime | None = None
    charge_cycles_count: int | None = None
    userfields: dict | None = None
    last_charged: datetime | None = None
    last_tracked_time: datetime | None = None
    next_estimated_charge_time: datetime | None = None

    @classmethod
    def from_current_response(cls, resp) -> Battery:
        return cls(
            id=resp.id,
            last_tracked_time=resp.last_tracked_time,
            next_estimated_charge_time=resp.next_estimated_charge_time,
        )

    @classmethod
    def from_details_response(cls, resp) -> Battery:
        return cls(
            id=resp.battery.id,
            name=resp.battery.name,
            description=resp.battery.description,
            used_in=resp.battery.used_in,
            charge_interval_days=resp.battery.charge_interval_days,
            created_timestamp=resp.battery.created_timestamp,
            userfields=resp.battery.userfields,
            charge_cycles_count=resp.charge_cycles_count,
            last_charged=resp.last_charged,
            last_tracked_time=resp.last_charged,
            next_estimated_charge_time=resp.next_estimated_charge_time,
        )

    def get_details(self, api_client):
        details = api_client.get_battery(self.id)
        if details:
            updated = Battery.from_details_response(details)
            self.__dict__.update(updated.__dict__)



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel


class Equipment(BaseModel):
    id: int
    name: str | None = None
    description: str | None = None
    instruction_manual_file_name: str | None = None
    created_timestamp: datetime | None = None
    userfields: dict | None = None

    @classmethod
    def from_equipment_data(cls, data) -> Equipment:
        return cls(
            id=data.id,
            name=data.name,
            description=data.description,
            instruction_manual_file_name=data.instruction_manual_file_name,
            created_timestamp=data.created_timestamp,
            userfields=data.userfields,
        )

    @classmethod
    def from_details_response(cls, resp) -> Equipment:
        return cls.from_equipment_data(resp.equipment)

    @classmethod
    def from_dict(cls, data: dict) -> Equipment:
        return cls(
            id=data.get("id"),
            name=data.get("name"),
            description=data.get("description"),
            instruction_manual_file_name=data.get("instruction_manual_file_name"),
            created_timestamp=data.get("row_created_timestamp")
            or data.get("created_timestamp"),
            userfields=data.get("userfields"),
        )

    def get_details(self, api_client):
        details = api_client.get_equipment(self.id)
        if details:
            updated = Equipment.from_details_response(details)
            self.__dict__.update(updated.__dict__)

    def as_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "instruction_manual_file_name": self.instruction_manual_file_name,
            "created_timestamp": self.created_timestamp,
            "userfields": self.userfields,
        }



from __future__ import annotations

from datetime import datetime
from enum import Enum

from pydantic import BaseModel

from .user import User


class PeriodType(str, Enum):
    MANUALLY = "manually"
    DYNAMIC_REGULAR = "dynamic-regular"
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    YEARLY = "yearly"
    ADAPTIVE = "adaptive"
    HOURLY = "hourly"


class AssignmentType(str, Enum):
    NO_ASSIGNMENT = "no-assignment"
    WHO_LEAST_DID_FIRST = "who-least-did-first"
    RANDOM = "random"
    IN_ALPHABETICAL_ORDER = "in-alphabetical-order"


class Chore(BaseModel):
    id: int
    name: str | None = None
    description: str | None = None
    period_type: PeriodType | None = None
    period_config: str | None = None
    period_days: int | None = None
    track_date_only: bool | None = None
    rollover: bool | None = None
    assignment_type: AssignmentType | None = None
    assignment_config: str | None = None
    next_execution_assigned_to_user_id: int | None = None
    userfields: dict | None = None
    last_tracked_time: datetime | None = None
    next_estimated_execution_time: datetime | None = None
    last_done_by: User | None = None
    track_count: int | None = None
    next_execution_assigned_user: User | None = None

    @classmethod
    def from_current_response(cls, resp) -> Chore:
        return cls(
            id=resp.chore_id,
            last_tracked_time=resp.last_tracked_time,
            next_estimated_execution_time=resp.next_estimated_execution_time,
        )

    @classmethod
    def from_details_response(cls, resp) -> Chore:
        chore_data = resp.chore
        period_type = (
            PeriodType(chore_data.period_type)
            if chore_data.period_type is not None
            else None
        )
        assignment_type = (
            AssignmentType(chore_data.assignment_type)
            if chore_data.assignment_type is not None
            else None
        )
        last_done_by = (
            User(
                id=resp.last_done_by.id,
                username=resp.last_done_by.username,
                first_name=resp.last_done_by.first_name,
                last_name=resp.last_done_by.last_name,
                display_name=resp.last_done_by.display_name,
            )
            if resp.last_done_by
            else None
        )
        next_user = (
            User(
                id=resp.next_execution_assigned_user.id,
                username=resp.next_execution_assigned_user.username,
                first_name=resp.next_execution_assigned_user.first_name,
                last_name=resp.next_execution_assigned_user.last_name,
                display_name=resp.next_execution_assigned_user.display_name,
            )
            if resp.next_execution_assigned_user
            else None
        )
        return cls(
            id=chore_data.id,
            name=chore_data.name,
            description=chore_data.description,
            period_type=period_type,
            period_config=chore_data.period_config,
            period_days=chore_data.period_days,
            track_date_only=chore_data.track_date_only,
            rollover=chore_data.rollover,
            assignment_type=assignment_type,
            assignment_config=chore_data.assignment_config,
            next_execution_assigned_to_user_id=chore_data.next_execution_assigned_to_user_id,
            userfields=chore_data.userfields,
            last_tracked_time=resp.last_tracked,
            next_estimated_execution_time=resp.next_estimated_execution_time,
            last_done_by=last_done_by,
            track_count=resp.track_count,
            next_execution_assigned_user=next_user,
        )

    def get_details(self, api_client):
        details = api_client.get_chore(self.id)
        if details:
            updated = Chore.from_details_response(details)
            self.__dict__.update(updated.__dict__)
            self.userfields = api_client.get_userfields("chores", self.id)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from ..data_models.chore import Chore

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class ChoreManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[Chore]:
        raw_chores = self._api.get_chores(query_filters)
        chores = [Chore.from_current_response(chore) for chore in raw_chores]
        if get_details:
            for chore in chores:
                chore.get_details(self._api)
        return chores

    def get(self, chore_id: int) -> Chore:
        resp = self._api.get_chore(chore_id)
        return Chore.from_details_response(resp)

    def execute(
        self,
        chore_id: int,
        done_by: int | None = None,
        tracked_time: datetime | None = None,
        skipped: bool = False,
    ):
        return self._api.execute_chore(chore_id, done_by, tracked_time, skipped)

    def undo(self, execution_id: int):
        return self._api.undo_chore_execution(execution_id)

    def merge(self, chore_id_keep: int, chore_id_remove: int):
        return self._api.merge_chores(chore_id_keep, chore_id_remove)

    def calculate_next_assignments(self):
        return self._api.calculate_chore_assignments()



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.equipment import Equipment
from ..data_models.generic import EntityType
from ..grocy_api_client import EquipmentData

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class EquipmentManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(
        self, query_filters: list[str] | None = None, get_details: bool = False
    ) -> list[Equipment]:
        raw_equipment = self._api.get_all_equipment(query_filters)
        equipment_items = [Equipment.from_dict(item) for item in raw_equipment]
        if get_details:
            for item in equipment_items:
                item.get_details(self._api)
        return equipment_items

    def get(self, equipment_id: int) -> Equipment | None:
        equipment_data = self._api.get_equipment(equipment_id)
        if equipment_data:
            return Equipment.from_details_response(equipment_data)
        return None

    def get_by_name(self, name: str) -> Equipment | None:
        query_filters = [f"name={name}"]
        equipment_items = self.list(query_filters, True)
        if equipment_items and len(equipment_items) > 0:
            return equipment_items[0]
        return None

    def get_all_objects(self) -> list[Equipment]:
        raw_equipment = self._api.get_generic_objects_for_type(EntityType.EQUIPMENT)
        if not raw_equipment:
            return []
        equipment_data = [EquipmentData(**eq) for eq in raw_equipment]
        equipment_items = []
        for item in equipment_data:
            equipment_details = self._api.get_equipment(item.id)
            if equipment_details:
                equipment_items.append(
                    Equipment.from_details_response(equipment_details)
                )
        return equipment_items



"""Public package interface for grocy."""

from .grocy import Grocy
from .managers import (
    BatteryManager,
    CalendarManager,
    ChoreManager,
    EquipmentManager,
    FileManager,
    GenericEntityManager,
    MealPlanManager,
    RecipeManager,
    ShoppingListManager,
    StockManager,
    SystemManager,
    TaskManager,
    UserManager,
)

__all__ = [
    "Grocy",
    "BatteryManager",
    "CalendarManager",
    "ChoreManager",
    "EquipmentManager",
    "FileManager",
    "GenericEntityManager",
    "MealPlanManager",
    "RecipeManager",
    "ShoppingListManager",
    "StockManager",
    "SystemManager",
    "TaskManager",
    "UserManager",
]

name = "grocy"



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel


class ProductBarcode(BaseModel):
    barcode: str
    amount: float | None = None


class QuantityUnit(BaseModel):
    id: int
    name: str
    name_plural: str | None = None
    description: str | None = None


class Product(BaseModel):
    id: int
    name: str | None = None
    product_group_id: int | None = None
    available_amount: float | None = None
    amount_aggregated: float | None = None
    amount_opened: float | None = None
    amount_opened_aggregated: float | None = None
    is_aggregated_amount: bool | None = None
    best_before_date: datetime | None = None
    barcodes: list[str] = []
    product_barcodes: list[ProductBarcode] = []
    amount_missing: float | None = None
    is_partly_in_stock: bool | None = None
    default_quantity_unit_purchase: QuantityUnit | None = None

    @classmethod
    def from_stock_response(cls, resp) -> Product:
        barcodes_list: list[ProductBarcode] = []
        name = None
        product_group_id = None
        if resp.product:
            name = resp.product.name
            product_group_id = resp.product.product_group_id
        return cls(
            id=resp.product_id,
            name=name,
            product_group_id=product_group_id,
            available_amount=resp.amount,
            amount_aggregated=resp.amount_aggregated,
            amount_opened=resp.amount_opened,
            amount_opened_aggregated=resp.amount_opened_aggregated,
            is_aggregated_amount=resp.is_aggregated_amount,
            best_before_date=resp.best_before_date,
            product_barcodes=barcodes_list,
            barcodes=[],
        )

    @classmethod
    def from_missing_response(cls, resp) -> Product:
        return cls(
            id=resp.id,
            name=resp.name,
            amount_missing=resp.amount_missing,
            is_partly_in_stock=resp.is_partly_in_stock,
        )

    @classmethod
    def from_details_response(cls, resp) -> Product:
        product_barcodes = (
            [ProductBarcode(barcode=b.barcode, amount=b.amount) for b in resp.barcodes]
            if resp.barcodes
            else []
        )
        barcodes = [b.barcode for b in product_barcodes]
        name = None
        product_group_id = None
        product_id = 0
        if resp.product:
            product_id = resp.product.id
            name = resp.product.name
            product_group_id = resp.product.product_group_id
        qu = None
        if resp.default_quantity_unit_purchase:
            qu = QuantityUnit(
                id=resp.default_quantity_unit_purchase.id,
                name=resp.default_quantity_unit_purchase.name,
                name_plural=resp.default_quantity_unit_purchase.name_plural,
                description=resp.default_quantity_unit_purchase.description,
            )
        return cls(
            id=product_id,
            name=name,
            product_group_id=product_group_id,
            available_amount=resp.stock_amount,
            best_before_date=resp.next_best_before_date,
            product_barcodes=product_barcodes,
            barcodes=barcodes,
            default_quantity_unit_purchase=qu,
        )

    @classmethod
    def from_product_data(cls, data) -> Product:
        return cls(
            id=data.id,
            name=data.name,
            product_group_id=data.product_group_id,
        )

    @classmethod
    def from_stock_log_response(cls, resp) -> Product:
        return cls(id=resp.product_id)

    def get_details(self, api_client):
        details = api_client.get_product(self.id)
        if details:
            updated = Product.from_details_response(details)
            # Preserve fields that the details response doesn't provide
            preserved = {}
            for field in ("amount_missing", "is_partly_in_stock"):
                current = getattr(self, field, None)
                if current is not None:
                    preserved[field] = current
            self.__dict__.update(updated.__dict__)
            for field, value in preserved.items():
                setattr(self, field, value)


class Group(BaseModel):
    id: int
    name: str
    description: str | None = None


class ShoppingListProduct(BaseModel):
    id: int
    product_id: int | None = None
    amount: float | None = None
    note: str | None = None
    product: Product | None = None
    done: bool = False

    @classmethod
    def from_shopping_list_item(cls, item) -> ShoppingListProduct:
        return cls(
            id=item.id,
            product_id=item.product_id,
            note=item.note,
            amount=item.amount,
            done=bool(item.done),
        )

    def get_details(self, api_client):
        if self.product_id:
            resp = api_client.get_product(self.product_id)
            if resp:
                self.product = Product.from_details_response(resp)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class CalendarManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def ical(self) -> str | None:
        return self._api.get_calendar_ical()

    def sharing_link(self) -> str | None:
        return self._api.get_calendar_sharing_link()



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel

from .user import User


class TaskCategory(BaseModel):
    id: int
    name: str
    description: str | None = None
    row_created_timestamp: datetime | None = None


class Task(BaseModel):
    id: int
    name: str
    description: str | None = None
    due_date: datetime | None = None
    done: int = 0
    done_timestamp: datetime | None = None
    category_id: int | None = None
    category: TaskCategory | None = None
    assigned_to_user_id: int | None = None
    assigned_to_user: User | None = None
    userfields: dict | None = None

    @classmethod
    def from_response(cls, resp) -> Task:
        category = None
        if resp.category:
            category = TaskCategory(
                id=resp.category.id,
                name=resp.category.name,
                description=resp.category.description,
                row_created_timestamp=resp.category.row_created_timestamp,
            )
        assigned_user = None
        if resp.assigned_to_user:
            assigned_user = User(
                id=resp.assigned_to_user.id,
                username=resp.assigned_to_user.username,
                first_name=resp.assigned_to_user.first_name,
                last_name=resp.assigned_to_user.last_name,
                display_name=resp.assigned_to_user.display_name,
            )
        return cls(
            id=resp.id,
            name=resp.name,
            description=resp.description,
            due_date=resp.due_date,
            done=resp.done,
            done_timestamp=resp.done_timestamp,
            category_id=resp.category_id,
            category=category,
            assigned_to_user_id=resp.assigned_to_user_id,
            assigned_to_user=assigned_user,
            userfields=resp.userfields,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.user import User

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class UserManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(self) -> list[User]:
        user_dtos = self._api.get_users()
        return [
            User(
                id=u.id,
                username=u.username,
                first_name=u.first_name,
                last_name=u.last_name,
                display_name=u.display_name,
            )
            for u in user_dtos
        ]

    def get(self, user_id: int) -> User | None:
        user = self._api.get_user(user_id=user_id)
        if user:
            return User(
                id=user.id,
                username=user.username,
                first_name=user.first_name,
                last_name=user.last_name,
                display_name=user.display_name,
            )
        return None

    def current(self) -> User | None:
        user = self._api.get_current_user()
        if user:
            return User(
                id=user.id,
                username=user.username,
                first_name=user.first_name,
                last_name=user.last_name,
                display_name=user.display_name,
            )
        return None

    def create(self, data: dict):
        return self._api.create_user(data)

    def edit(self, user_id: int, data: dict):
        return self._api.edit_user(user_id, data)

    def delete(self, user_id: int):
        return self._api.delete_user(user_id)

    def settings(self):
        return self._api.get_user_settings()

    def get_setting(self, key: str):
        return self._api.get_user_setting(key)

    def set_setting(self, key: str, value):
        return self._api.set_user_setting(key, value)

    def delete_setting(self, key: str):
        return self._api.delete_user_setting(key)



from .batteries import BatteryManager
from .calendar import CalendarManager
from .chores import ChoreManager
from .equipment import EquipmentManager
from .files import FileManager
from .generic import GenericEntityManager
from .meal_plan import MealPlanManager
from .recipes import RecipeManager
from .shopping_list import ShoppingListManager
from .stock import StockManager
from .system import SystemManager
from .tasks import TaskManager
from .users import UserManager

__all__ = [
    "BatteryManager",
    "CalendarManager",
    "ChoreManager",
    "EquipmentManager",
    "FileManager",
    "GenericEntityManager",
    "MealPlanManager",
    "RecipeManager",
    "ShoppingListManager",
    "StockManager",
    "SystemManager",
    "TaskManager",
    "UserManager",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.meal_items import RecipeItem

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class RecipeManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def get(self, recipe_id: int) -> RecipeItem | None:
        recipe = self._api.get_recipe(recipe_id)
        if recipe:
            return RecipeItem.from_response(recipe)
        return None

    def consume(self, recipe_id: int):
        return self._api.consume_recipe(recipe_id)

    def fulfillment(self, recipe_id: int):
        return self._api.get_recipe_fulfillment(recipe_id)

    def all_fulfillment(self):
        return self._api.get_all_recipes_fulfillment()

    def copy(self, recipe_id: int):
        return self._api.copy_recipe(recipe_id)

    def add_not_fulfilled_to_shopping_list(self, recipe_id: int):
        return self._api.add_not_fulfilled_to_shopping_list(recipe_id)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.product import ShoppingListProduct

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class ShoppingListManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def items(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[ShoppingListProduct]:
        raw = self._api.get_shopping_list(query_filters)
        shopping_list = [
            ShoppingListProduct.from_shopping_list_item(item) for item in raw
        ]
        if get_details:
            for item in shopping_list:
                item.get_details(self._api)
        return shopping_list

    def add_product(
        self,
        product_id: int,
        shopping_list_id: int | None = None,
        amount: float | None = None,
        quantity_unit_id: int | None = None,
    ):
        return self._api.add_product_to_shopping_list(
            product_id, shopping_list_id, amount, quantity_unit_id
        )

    def remove_product(
        self, product_id: int, shopping_list_id: int = 1, amount: float = 1
    ):
        return self._api.remove_product_in_shopping_list(
            product_id, shopping_list_id, amount
        )

    def clear(self, shopping_list_id: int = 1):
        return self._api.clear_shopping_list(shopping_list_id)

    def add_missing_products(self, shopping_list_id: int = 1):
        return self._api.add_missing_product_to_shopping_list(shopping_list_id)

    def add_overdue_products(self, shopping_list_id: int = 1):
        return self._api.add_overdue_products_to_shopping_list(shopping_list_id)

    def add_expired_products(self, shopping_list_id: int = 1):
        return self._api.add_expired_products_to_shopping_list(shopping_list_id)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from ..data_models.task import Task

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class TaskManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(self, query_filters: list[str] | None = None) -> list[Task]:
        raw_tasks = self._api.get_tasks(query_filters)
        return [Task.from_response(task) for task in raw_tasks]

    def get(self, task_id: int) -> Task:
        resp = self._api.get_task(task_id)
        return Task.from_response(resp)

    def complete(self, task_id: int, done_time: datetime | None = None):
        return self._api.complete_task(task_id, done_time)

    def undo(self, task_id: int):
        return self._api.undo_task(task_id)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from ..data_models.generic import EntityType
from ..data_models.product import Group, Product
from ..grocy_api_client import ProductData, TransactionType

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class StockManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def current(self) -> list[Product]:
        raw_stock = self._api.get_stock()
        return [Product.from_stock_response(resp) for resp in raw_stock]

    def due_products(self, get_details: bool = False) -> list[Product]:
        raw = self._api.get_volatile_stock().due_products
        products = [Product.from_stock_response(resp) for resp in raw] if raw else []
        if get_details:
            for item in products:
                item.get_details(self._api)
        return products

    def overdue_products(self, get_details: bool = False) -> list[Product]:
        raw = self._api.get_volatile_stock().overdue_products
        products = [Product.from_stock_response(resp) for resp in raw] if raw else []
        if get_details:
            for item in products:
                item.get_details(self._api)
        return products

    def expired_products(self, get_details: bool = False) -> list[Product]:
        raw = self._api.get_volatile_stock().expired_products
        products = [Product.from_stock_response(resp) for resp in raw] if raw else []
        if get_details:
            for item in products:
                item.get_details(self._api)
        return products

    def missing_products(self, get_details: bool = False) -> list[Product]:
        raw = self._api.get_volatile_stock().missing_products
        products = [Product.from_missing_response(resp) for resp in raw] if raw else []
        if get_details:
            for item in products:
                item.get_details(self._api)
        return products

    def product(self, product_id: int) -> Product | None:
        resp = self._api.get_product(product_id)
        if resp:
            return Product.from_details_response(resp)
        return None

    def product_by_barcode(self, barcode: str) -> Product | None:
        resp = self._api.get_product_by_barcode(barcode)
        if resp:
            return Product.from_details_response(resp)
        return None

    def all_products(self) -> list[Product]:
        raw_products = self._api.get_generic_objects_for_type(EntityType.PRODUCTS)
        if not raw_products:
            return []
        product_datas = [ProductData(**product) for product in raw_products]
        return [Product.from_product_data(product) for product in product_datas]

    def add(
        self,
        product_id: int,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
        transaction_type: TransactionType = TransactionType.PURCHASE,
    ):
        return self._api.add_product(
            product_id, amount, price, best_before_date, transaction_type
        )

    def consume(
        self,
        product_id: int,
        amount: float = 1,
        spoiled: bool = False,
        transaction_type: TransactionType = TransactionType.CONSUME,
        allow_subproduct_substitution: bool = False,
    ):
        return self._api.consume_product(
            product_id, amount, spoiled, transaction_type, allow_subproduct_substitution
        )

    def open(
        self,
        product_id: int,
        amount: float = 1,
        allow_subproduct_substitution: bool = False,
    ):
        return self._api.open_product(product_id, amount, allow_subproduct_substitution)

    def transfer(
        self,
        product_id: int,
        amount: float,
        location_from: int,
        location_to: int,
    ):
        return self._api.transfer_product(
            product_id, amount, location_from, location_to
        )

    def inventory(
        self,
        product_id: int,
        new_amount: float,
        best_before_date: datetime | None = None,
        shopping_location_id: int | None = None,
        location_id: int | None = None,
        price: float | None = None,
        get_details: bool = True,
    ) -> Product | None:
        resp = self._api.inventory_product(
            product_id,
            new_amount,
            best_before_date,
            shopping_location_id,
            location_id,
            price,
        )
        if resp:
            product = Product.from_stock_log_response(resp)
            if get_details:
                product.get_details(self._api)
            return product
        return None

    def add_by_barcode(
        self,
        barcode: str,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
        get_details: bool = True,
    ) -> Product | None:
        resp = self._api.add_product_by_barcode(
            barcode, amount, price, best_before_date
        )
        if resp:
            product = Product.from_stock_log_response(resp)
            if get_details:
                product.get_details(self._api)
            return product
        return None

    def consume_by_barcode(
        self,
        barcode: str,
        amount: float = 1,
        spoiled: bool = False,
        get_details: bool = True,
    ) -> Product | None:
        resp = self._api.consume_product_by_barcode(barcode, amount, spoiled)
        if resp:
            product = Product.from_stock_log_response(resp)
            if get_details:
                product.get_details(self._api)
            return product
        return None

    def transfer_by_barcode(
        self,
        barcode: str,
        amount: float,
        location_from: int,
        location_to: int,
    ):
        return self._api.transfer_product_by_barcode(
            barcode, amount, location_from, location_to
        )

    def open_by_barcode(self, barcode: str, amount: float = 1):
        return self._api.open_product_by_barcode(barcode, amount)

    def inventory_by_barcode(
        self,
        barcode: str,
        new_amount: float,
        best_before_date: datetime | None = None,
        location_id: int | None = None,
        price: float | None = None,
        get_details: bool = True,
    ) -> Product | None:
        resp = self._api.inventory_product_by_barcode(
            barcode, new_amount, best_before_date, location_id, price
        )
        if resp:
            product = Product.from_stock_log_response(resp)
            if get_details:
                product.get_details(self._api)
            return product
        return None

    def merge(self, product_id_keep: int, product_id_remove: int):
        return self._api.merge_products(product_id_keep, product_id_remove)

    def entry(self, entry_id: int):
        return self._api.get_stock_entry(entry_id)

    def edit_entry(self, entry_id: int, data: dict):
        return self._api.edit_stock_entry(entry_id, data)

    def product_entries(self, product_id: int):
        return self._api.get_product_stock_entries(product_id)

    def product_locations(self, product_id: int):
        return self._api.get_product_stock_locations(product_id)

    def product_price_history(self, product_id: int):
        return self._api.get_product_price_history(product_id)

    def entries_by_location(self, location_id: int):
        return self._api.get_stock_entries_by_location(location_id)

    def booking(self, booking_id: int):
        return self._api.get_stock_booking(booking_id)

    def undo_booking(self, booking_id: int):
        return self._api.undo_stock_booking(booking_id)

    def transaction(self, transaction_id: str):
        return self._api.get_stock_transactions(transaction_id)

    def undo_transaction(self, transaction_id: str):
        return self._api.undo_stock_transaction(transaction_id)

    def barcode_lookup(self, barcode: str):
        return self._api.external_barcode_lookup(barcode)

    def product_groups(self, query_filters: list[str] | None = None) -> list[Group]:
        raw_groups = self._api.get_product_groups(query_filters)
        return [
            Group(id=resp.id, name=resp.name, description=resp.description)
            for resp in raw_groups
        ]

    def upload_product_picture(self, product_id: int, pic_path: str):
        self._api.upload_product_picture(product_id, pic_path)
        return self._api.update_product_pic(product_id)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.generic import EntityType

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class GenericEntityManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(
        self, entity_type: EntityType, query_filters: list[str] | None = None
    ) -> list[dict]:
        return (
            self._api.get_generic_objects_for_type(entity_type.value, query_filters)
            or []
        )

    def get(self, entity_type: EntityType, object_id: int) -> dict:
        return self._api.get_generic(entity_type.value, object_id)

    def create(self, entity_type: EntityType, data) -> dict:
        return self._api.add_generic(entity_type.value, data)

    def update(self, entity_type: EntityType, object_id: int, data):
        return self._api.update_generic(entity_type.value, object_id, data)

    def delete(self, entity_type: EntityType, object_id: int):
        return self._api.delete_generic(entity_type.value, object_id)

    def get_userfields(self, entity: str, object_id: int):
        return self._api.get_userfields(entity, object_id)

    def set_userfields(self, entity: str, object_id: int, key: str, value):
        return self._api.set_userfields(entity, object_id, key, value)



from __future__ import annotations

import base64
from datetime import date, datetime
from enum import Enum

from pydantic import BaseModel


class RecipeItem(BaseModel):
    id: int | None = None
    name: str
    description: str | None = None
    base_servings: int
    desired_servings: int
    picture_file_name: str | None = None

    @classmethod
    def from_response(cls, resp) -> RecipeItem:
        return cls(
            id=resp.id,
            name=resp.name,
            description=resp.description,
            base_servings=resp.base_servings,
            desired_servings=resp.desired_servings,
            picture_file_name=resp.picture_file_name,
        )

    def get_picture_url_path(self, width: int = 400):
        if self.picture_file_name:
            b64name = base64.b64encode(self.picture_file_name.encode("ascii"))
            path = "files/recipepictures/" + str(b64name, "utf-8")
            return f"{path}?force_serve_as=picture&best_fit_width={width}"


class MealPlanSection(BaseModel):
    id: int | None = None
    name: str | None = None
    sort_number: int | None = None
    row_created_timestamp: datetime | None = None

    @classmethod
    def from_response(cls, resp) -> MealPlanSection:
        return cls(
            id=resp.id,
            name=resp.name,
            sort_number=resp.sort_number,
            row_created_timestamp=resp.row_created_timestamp,
        )


class MealPlanItemType(str, Enum):
    NOTE = "note"
    PRODUCT = "product"
    RECIPE = "recipe"


class MealPlanItem(BaseModel):
    id: int
    day: date | None = None
    recipe_id: int | None = None
    recipe_servings: int | None = None
    note: str | None = None
    recipe: RecipeItem | None = None
    section_id: int | None = None
    section: MealPlanSection | None = None
    type: MealPlanItemType
    product_id: int | None = None

    @classmethod
    def from_response(cls, resp) -> MealPlanItem:
        day = cls._normalize_day(resp.day)
        return cls(
            id=resp.id,
            day=day,
            recipe_id=resp.recipe_id,
            recipe_servings=resp.recipe_servings,
            note=resp.note,
            section_id=resp.section_id,
            type=MealPlanItemType(resp.type),
            product_id=resp.product_id,
        )

    def get_details(self, api_client):
        if self.recipe_id:
            recipe = api_client.get_recipe(self.recipe_id)
            if recipe:
                self.recipe = RecipeItem.from_response(recipe)
        if self.section_id:
            section = api_client.get_meal_plan_section(self.section_id)
            if section:
                self.section = MealPlanSection.from_response(section)

    @staticmethod
    def _normalize_day(value) -> date | None:
        if value is None:
            return None
        if isinstance(value, date) and not isinstance(value, datetime):
            return value
        if isinstance(value, datetime):
            return value.date()
        return datetime.fromisoformat(str(value)).date()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from ..data_models.system import SystemConfig, SystemInfo, SystemTime

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class SystemManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def info(self) -> SystemInfo | None:
        raw = self._api.get_system_info()
        if raw:
            return SystemInfo.from_dto(raw)
        return None

    def time(self) -> SystemTime | None:
        raw = self._api.get_system_time()
        if raw:
            return SystemTime.from_dto(raw)
        return None

    def config(self) -> SystemConfig | None:
        raw = self._api.get_system_config()
        if raw:
            return SystemConfig.from_dto(raw)
        return None

    def db_changed_time(self) -> datetime | None:
        return self._api.get_last_db_changed()



from __future__ import annotations

from typing import TYPE_CHECKING

from ..data_models.meal_items import MealPlanItem, MealPlanSection

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class MealPlanManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def items(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[MealPlanItem]:
        raw = self._api.get_meal_plan(query_filters)
        meal_plan = [MealPlanItem.from_response(data) for data in raw]
        if get_details:
            for item in meal_plan:
                item.get_details(self._api)
        return meal_plan

    def sections(self, query_filters: list[str] | None = None) -> list[MealPlanSection]:
        raw = self._api.get_meal_plan_sections(query_filters)
        return [MealPlanSection.from_response(section) for section in raw]

    def section(self, section_id: int) -> MealPlanSection | None:
        section = self._api.get_meal_plan_section(section_id)
        if section:
            return MealPlanSection.from_response(section)
        return None



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from ..data_models.battery import Battery

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class BatteryManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def list(
        self, query_filters: list[str] | None = None, get_details: bool = False
    ) -> list[Battery]:
        raw = self._api.get_batteries(query_filters)
        batteries = [Battery.from_current_response(bat) for bat in raw]
        if get_details:
            for item in batteries:
                item.get_details(self._api)
        return batteries

    def get(self, battery_id: int) -> Battery | None:
        battery = self._api.get_battery(battery_id)
        if battery:
            return Battery.from_details_response(battery)
        return None

    def charge(self, battery_id: int, tracked_time: datetime | None = None):
        return self._api.charge_battery(battery_id, tracked_time)

    def undo(self, charge_cycle_id: int):
        return self._api.undo_battery_charge(charge_cycle_id)



from datetime import date, datetime

from pydantic import BaseModel


class SystemInfo(BaseModel):
    grocy_version: str
    grocy_release_date: date
    php_version: str
    sqlite_version: str
    os: str
    client: str

    @classmethod
    def from_dto(cls, dto) -> "SystemInfo":
        return cls(
            grocy_version=dto.grocy_version_info.version,
            grocy_release_date=dto.grocy_version_info.release_date,
            php_version=dto.php_version,
            sqlite_version=dto.sqlite_version,
            os=dto.os,
            client=dto.client,
        )


class SystemTime(BaseModel):
    timezone: str
    time_local: datetime
    time_local_sqlite3: datetime
    time_utc: datetime
    timestamp: int

    @classmethod
    def from_dto(cls, dto) -> "SystemTime":
        return cls(
            timezone=dto.timezone,
            time_local=dto.time_local,
            time_local_sqlite3=dto.time_local_sqlite3,
            time_utc=dto.time_utc,
            timestamp=dto.timestamp,
        )


class SystemConfig(BaseModel):
    username: str
    base_path: str
    base_url: str
    mode: str
    default_locale: str
    locale: str
    currency: str
    enabled_features: list[str] = []

    @classmethod
    def from_dto(cls, dto) -> "SystemConfig":
        enabled_features = [
            feature
            for feature, value in dto.feature_flags.items()
            if value not in (False, "0")
        ]
        return cls(
            username=dto.username,
            base_path=dto.base_path,
            base_url=dto.base_url,
            mode=dto.mode,
            default_locale=dto.default_locale,
            locale=dto.locale,
            currency=dto.currency,
            enabled_features=enabled_features,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..grocy_api_client import GrocyApiClient


class FileManager:
    def __init__(self, api_client: GrocyApiClient):
        self._api = api_client

    def upload(self, group: str, file_name: str, data):
        return self._api.upload_file(group, file_name, data)

    def download(self, group: str, file_name: str):
        return self._api.download_file(group, file_name)

    def delete(self, group: str, file_name: str):
        return self._api.delete_file(group, file_name)
